#include "lcd/lcd.h"
#include "fatfs/tf_card.h"
#include <string.h>
#include "gd32vf103_i2c.h"

//#define HZ2MS(X) ((uint64_t)((X * 53000.0)/(SystemCoreClock)))


void init_uart0(void)
{	
	/* enable GPIO clock */
    rcu_periph_clock_enable(RCU_GPIOA);
    /* enable USART clock */
    rcu_periph_clock_enable(RCU_USART0);

    /* connect port to USARTx_Tx */
    gpio_init(GPIOA, GPIO_MODE_AF_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_9);
    /* connect port to USARTx_Rx */
    gpio_init(GPIOA, GPIO_MODE_IN_FLOATING, GPIO_OSPEED_50MHZ, GPIO_PIN_10);

	/* USART configure */
    usart_deinit(USART0);
    usart_baudrate_set(USART0, 115200U);
    usart_word_length_set(USART0, USART_WL_8BIT);
    usart_stop_bit_set(USART0, USART_STB_1BIT);
    usart_parity_config(USART0, USART_PM_NONE);
    usart_hardware_flow_rts_config(USART0, USART_RTS_DISABLE);
    usart_hardware_flow_cts_config(USART0, USART_CTS_DISABLE);
    usart_receive_config(USART0, USART_RECEIVE_ENABLE);
    usart_transmit_config(USART0, USART_TRANSMIT_ENABLE);
    usart_enable(USART0);

    usart_interrupt_enable(USART0, USART_INT_RBNE);
}


#define MAX_ROW 20
#define MAX_NMEA 192

static inline void i2c_wait_busy(void)
{
    while(i2c_flag_get(I2C0, I2C_FLAG_I2CBSY))
            ;
}

static void init_i2c0(void)
{
    rcu_periph_clock_enable(RCU_GPIOB);
    rcu_periph_clock_enable(RCU_I2C0);
    gpio_init(GPIOB, GPIO_MODE_AF_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_6 | GPIO_PIN_7);
    //gpio_pin_remap_config(GPIO_I2C0_REMAP, ENABLE);
    i2c_deinit(I2C0);
    i2c_clock_config(I2C0, 100000, I2C_DTCY_2);
    i2c_mode_addr_config(I2C0, I2C_I2CMODE_ENABLE, I2C_ADDFORMAT_7BITS, 0);
    i2c_enable(I2C0);
    i2c_ack_config(I2C0, I2C_ACK_ENABLE);
}


#define BMP180_I2C_ADDR 0xEE

/* Oversapling setting (0 = Ultra low pow, 3 = Ultra high res) */
#define BMP180_OSS_ULP 0
#define BMP180_OSS_STD 1
#define BMP180_OSS_HR  2
#define BMP180_OSS_UHR 3

#define BMP180_REG_T        0x2E
static const uint32_t BMP180_OSS = BMP180_OSS_STD; 
static const uint8_t BMP180_REG_P_ARRAY[4] = {0x34, 0x74, 0xB4, 0xF4};
static const uint8_t BMP180_REG_P_TIME[4] = {5, 8, 14, 26};
#define BMP180_REG_P BMP180_REG_P_ARRAY[BMP180_OSS]
#define BMP180_P_CONVERSION_TIME BMP180_REG_P_TIME[BMP180_OSS_STD]

struct __attribute__((packed)) bmp180_cal_regs {
    union {
        struct __attribute__((packed)) creg_s {
            /* Read: 22 B */
            int16_t AC1, AC2, AC3;
            uint16_t AC4, AC5, AC6;
            int16_t B1, B2, MB, MC, MD;
        } creg;
        uint16_t reg_ptr[11];
    };
    /* Calculated params */
    int32_t B3, B5, B6;
    uint32_t B4, B7;
};

static struct bmp180_cal_regs BMP180;

/* Temperature [in 0.1 °C] */
static int32_t compensate_temperature(int16_t ut)
{
    int32_t x1 = ((ut - BMP180.creg.AC6) * BMP180.creg.AC5) >> 15;
    int32_t x2 = (BMP180.creg.MC << 11) / (x1 + BMP180.creg.MD);
    BMP180.B5 = x1 + x2;
    return (BMP180.B5 + 8) >> 4;
}

/* Pressure [in Pa, 1 Pa = 0.01mbar] */
static uint32_t compensate_pressure(uint32_t up)
{
    uint32_t x1, x2, x3;
    uint32_t p;
    BMP180.B6 = BMP180.B5 - 4000;
    x1 = (BMP180.creg.B2 * ((BMP180.B6 * BMP180.B6) >> 12)) >> 11;
    x2 = (BMP180.creg.AC2 * BMP180.B6) >> 11;
    x3 = x1 + x2;
    BMP180.B3 = ((((int32_t)BMP180.creg.AC1 * 4 + x3) << BMP180_OSS) + 2) >> 2;
    x1 = (BMP180.creg.AC3 * BMP180.B6) >> 13;
    x2 = (BMP180.creg.B1 * ((BMP180.B6 * BMP180.B6) >> 12)) >> 16;
    x3 = ((x1 + x2) + 2)  >> 2;
    BMP180.B4 = (BMP180.creg.AC4 * (x3 + 32768)) >> 15; 
    BMP180.B7 = (up - BMP180.B3) * (50000 >> BMP180_OSS);
    if (BMP180.B7 < 0x80000000)
        p = (BMP180.B7 * 2) / BMP180.B4;
    else
        p = (BMP180.B7 / BMP180.B4) * 2;
    x1 = (p >> 8) * (p >> 8);
    x1 = (x1 * 3038) >> 16;
    x2 = (-7357 * p) >> 16;
    p += (x1 + x2 + 3791) >> 4;
    return p;
}

static void bmp180_calibrate(void)
{
    uint8_t reg_msb, reg_lsb;
    uint8_t r_start = 0xAA;
    unsigned i;
    i2c_wait_busy();
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_TRANSMITTER);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, 0xAA);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_RECEIVER);
    i2c_ack_config(I2C0, I2C_ACK_ENABLE);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    for (i = 0; i < 10; i++) {
        while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
            ;
        reg_msb = i2c_data_receive(I2C0);
        while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
            ;
        reg_lsb = i2c_data_receive(I2C0);
        BMP180.reg_ptr[i] = (reg_msb << 8) + reg_lsb;
    }
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    reg_msb = i2c_data_receive(I2C0);
    while(!i2c_flag_get(I2C0, I2C_FLAG_BTC))
        ;
    i2c_ack_config(I2C0, I2C_ACK_DISABLE);
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    reg_lsb = i2c_data_receive(I2C0);
    i2c_stop_on_bus(I2C0);
    while (I2C_CTL0(I2C0) & 0x0200)
        ;
    BMP180.reg_ptr[10] = (reg_msb << 8) + reg_lsb;
}

static int16_t bmp180_read_T(void)
{
    uint8_t data_msb, data_lsb;
    i2c_wait_busy();
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_TRANSMITTER);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, 0xF4);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, BMP180_REG_T);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_stop_on_bus(I2C0);
    while (I2C_CTL0(I2C0) & 0x0200)
        ;

    delay_1ms(5);
    i2c_wait_busy();
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_TRANSMITTER);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, 0xF6);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_RECEIVER);
    i2c_ack_config(I2C0, I2C_ACK_ENABLE);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    data_msb = i2c_data_receive(I2C0);
    while(!i2c_flag_get(I2C0, I2C_FLAG_BTC))
        ;
    i2c_ack_config(I2C0, I2C_ACK_DISABLE);
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    data_lsb = i2c_data_receive(I2C0);
    i2c_stop_on_bus(I2C0);
    while (I2C_CTL0(I2C0) & 0x0200)
        ;
    return (int16_t)(data_lsb + (data_msb << 8));
}

static uint32_t bmp180_read_P(void)
{
    uint8_t data_msb, data_lsb, data_xlsb;
    i2c_wait_busy();
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_TRANSMITTER);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, 0xF4);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, BMP180_REG_P);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_stop_on_bus(I2C0);
    while (I2C_CTL0(I2C0) & 0x0200)
        ;

    delay_1ms(5);
    i2c_wait_busy();
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_TRANSMITTER);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_data_transmit(I2C0, 0xF6);
    while (!i2c_flag_get(I2C0, I2C_FLAG_TBE))
        ;
    i2c_start_on_bus(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_SBSEND))
        ;
    i2c_master_addressing(I2C0, BMP180_I2C_ADDR, I2C_RECEIVER);
    i2c_ack_config(I2C0, I2C_ACK_ENABLE);
    while (!i2c_flag_get(I2C0, I2C_FLAG_ADDSEND))
        ;
    i2c_flag_clear(I2C0, I2C_FLAG_ADDSEND);
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    data_msb = i2c_data_receive(I2C0);
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    data_lsb = i2c_data_receive(I2C0);
    while(!i2c_flag_get(I2C0, I2C_FLAG_BTC))
        ;
    i2c_ack_config(I2C0, I2C_ACK_DISABLE);
    while (!i2c_flag_get(I2C0, I2C_FLAG_RBNE))
        ;
    data_xlsb = i2c_data_receive(I2C0);
    i2c_stop_on_bus(I2C0);
    while (I2C_CTL0(I2C0) & 0x0200)
        ;
    return (uint32_t)((data_xlsb + (data_lsb << 8) + (data_msb << 16)) >> 3);
}

int main(void)
{
    uint8_t mount_is_ok = 1; /* 0: mount successful ; 1: mount failed */
    char dte[MAX_ROW]="Date";
    char gps[MAX_ROW]="GPS Data";
    char prs[MAX_ROW]="Pressure";
    char tmp[MAX_ROW]="Temperature";
    char nmea_buf[MAX_NMEA];
    int nmea_i = 0;
    int16_t ut;
    uint32_t up;
    uint32_t pressure;
    int32_t temperature;

    SystemInit();
    /* Init GPIOs */
    rcu_periph_clock_enable(RCU_GPIOA);
    rcu_periph_clock_enable(RCU_GPIOC);
    LEDR(0);

    /* User Button */
    gpio_init(GPIOA, GPIO_MODE_IN_FLOATING, GPIO_OSPEED_2MHZ, GPIO_PIN_8);

    /* GPS on/off */
    gpio_init(GPIOA, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, GPIO_PIN_0);
    gpio_bit_write(GPIOA, GPIO_PIN_0, 0);

    /* HW init: uart, ADC, Display */
    init_uart0();
    Lcd_Init();

    /* I2C init */
    init_i2c0();

    /* Read calibration registers */
    bmp180_calibrate();




    while (1)
    {
        ut = bmp180_read_T();
        up = bmp180_read_P();

        temperature = compensate_temperature(ut);
        pressure = compensate_pressure(up);
        snprintf(tmp, MAX_ROW - 1, "T: %d", temperature);
        snprintf(prs, MAX_ROW - 1, "p: %u", pressure);
        //snprintf(tmp, MAX_ROW - 1, "T: %d", ut);
        //snprintf(prs, MAX_ROW - 1, "p: %u", up);
        /* Display background */
        LCD_Clear(BLACK);
        BACK_COLOR=BLACK;
        LCD_ShowString (12, 16, (u8 *)dte, GREEN);
        LCD_ShowString(12, 32, (u8 *)(gps), GREEN);
        LCD_ShowString(12, 48, (u8 *)(prs), GREEN);
        LCD_ShowString(12, 64, (u8 *)(tmp), GREEN);
        delay_1ms(10000);
#if 0
        //nmea_buf[nmea_i] = usart_data_receive(USART1);
        if (nmea_buf[nmea_i] == 0)
            continue;
        nmea_i++;
        if (nmea_buf[nmea_i] != '\r')
            continue;
        //mingps_parse(nmea_buf...);
        nmea_i = 0;
#endif
    }
}

int _put_char(int ch)
{
    usart_data_transmit(USART0, (uint8_t) ch );
    while ( usart_flag_get(USART0, USART_FLAG_TBE)== RESET){
    }

    return ch;
}
